#include "MusicClock.h"

#include <avr/io.h>
#include <avr/wdt.h>
#include <stdio.h>
#include <string.h>
#include <avr/power.h>
#include <avr/cpufunc.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "pins.h"
#include "Lib/ff.h"
#include "Lib/diskio.h"
#include "timer.h"
#include "lcd.h"
#include "rtc.h"
#include "mp3.h"
#include "irctrl.h"

#if defined(__AVR_AT90USB1286__) && defined(USE_USB)
#include "Lib/MassStorage.h"
#else
// atmega644 or not using usb, make the USB related stuff do nothing
#define USB_DoTask() _NOP() // do nothing
#define USB_IsConnected() (0) // never connected nothing
#define USB_Init() _NOP() // do nothing
#define USB_Enable() _NOP() // do nothing
#define USB_Disable() _NOP() // do nothing
#endif

#define VOLUME_STEP 5

// to be stored on DS1307 RAM
// settings that change frequently but must be remembered on boot
static volatile struct
{
    int16_t volume; // current volume, 1 to 255, but signed and larger for fun
    bool_t mute;
    playmode_t playMode;
    submenu_t submenu;
    timemode_t timeMode;
}
ramSettings;

// to be stored in internal EEPROM
// settings that change rarely but must be remembered on boot
static volatile struct
{
    // reminder, day-of-week are indexed 0 to 6, monday to sunday
    uint8_t alarmMask; // each bit is alarm on or off for a particular day-of-week
    uint8_t alarmHour[7];
    uint8_t alarmMinute[7];
}
eepromSettings;

static volatile struct
{
    uint8_t dayOfWeek; // 0 to 6, monday to sunday
    uint8_t hour; // 1 to 24, never 0 please
    uint8_t minute;
    uint8_t second;
}
currentTime;

static volatile uint8_t oneShotCnt;
static volatile bool_t isPlaying;
static volatile bool_t isInMenu;
static volatile bool_t isUsbEnabled = bool_false;
static volatile bool_t alarmFlag;
static volatile bool_t snoozeFlag;
static volatile lcdcontent_t lcdContent;
static volatile bool_t updateRtc;
static volatile mp3_playstatus_t lastPlayStatus;

static volatile FRESULT err = -1;
static volatile bool_t success = bool_false;
static FATFS fatfs;
static DIR dir;
static FILINFO fileInfo;
static FIL file;

// stream to LCD so we can use printf
static int lcd_putchar(char c, FILE* stream);
static FILE lcd_stdout = FDEV_SETUP_STREAM(lcd_putchar, NULL, _FDEV_SETUP_WRITE);
static int lcd_putchar(char c, FILE* stream)
{
    lcd_sendChar(c);
}

void oneShotTimer_event()
{
    isInMenu = bool_false;
    lcdContent = lcdshow_currentTime;
    led_on();
}

void tmr0_oneSecondEvent()
{
    currentTime.second++;
    
    if (oneShotCnt == 0)
    {
        oneShotTimer_event();
    }
    else
    {
        oneShotCnt--;
    }

    if (currentTime.hour == 0)
    {
        currentTime.hour = 24;
    }
    
    if (currentTime.second >= 60)
    {
        currentTime.second = 0;
        currentTime.minute++;
        updateRtc = bool_true;
    }
    
    if (currentTime.minute >= 60)
    {
        currentTime.minute = 0;
        currentTime.hour++;
    }
    
    if (currentTime.hour > 24)
    {
        currentTime.hour = 1;
        currentTime.dayOfWeek++;
    }
    
    if (currentTime.dayOfWeek >= 7)
    {
        currentTime.dayOfWeek = 0;
    }
    
    if (bit_is_set(eepromSettings.alarmMask, currentTime.dayOfWeek) && // if enabled
        eepromSettings.alarmHour[currentTime.dayOfWeek] == currentTime.hour && // if match
        eepromSettings.alarmMinute[currentTime.dayOfWeek] == currentTime.minute) // if match
    {
        alarmFlag = bool_true;
    }
    else
    {
        alarmFlag = bool_false;
        snoozeFlag = bool_false;

        if (isPlaying == bool_false)
        {
            mp3_powerDown();
        }
    }
}

void oneShotTimer_go(uint8_t seconds)
{
    oneShotCnt = seconds;
}

mp3_playstatus_t mp3_playFile(FIL* file)
{
    if (bit_is_clear(MP3_PORTINPUT, MP3_PIN_DREQ)) return playstatus_busy;
    // the location of this may seem stupid, but it actually makes performance better
    // i have no idea why but adding this made stuttering go away
    
    while (bit_is_set(MP3_PORTINPUT, MP3_PIN_DREQ))
    {
        uint8_t buff[32];
        UINT read;
        err = f_read(file, buff, 32, &read); // read the file
        
        mp3_playPacket(buff, read); // play the sound
        
        if (read != 32) return playstatus_endOfFile;
        if (err != 0) return playstatus_endOfFile;
    }
    
    return playstatus_ok; // almost guaranteed to not reach this point
}

void init(void)
{
	// enable watchdog timer
	MCUSR &= ~(1 << WDRF);
    wdt_enable(WDTO_4S);

    led_output();
    led_on();
    
#if defined(__AVR_AT90USB1286__)
    // enable clock division since we are using 3.3V
	clock_prescale_set(clock_div_2);

	USB_Init();
#endif

	lcd_init();
	rtc_init();
	mp3_init();
	tmr0_init();
    irctrl_init();
	
	sei(); // enable interrupts

    wdt_reset();

    rtc_run();
	
	loadFromEeprom();
	loadFromRam();

	if (ramSettings.mute == bool_false)
	{
        mp3_setVolume(ramSettings.volume);
	}
    else
    {
        mp3_setVolume(1);
    }

    mp3_powerDown();

    lcdContent = lcdshow_currentTime;
    isUsbEnabled = bool_false;

    rtc_getTimeBurst(0, 0, 0, &(currentTime.dayOfWeek), &(currentTime.hour), &(currentTime.minute), &(currentTime.second));
    currentTime.dayOfWeek--; // rtc reports from 1-7, we want 0-6
}

#include "menu.c"
#include "settings.c"
#include "songnav.c"
#include "alarm.c"
#include "tests.c"

int main(void)
{
	init();

    wdt_reset();
	
	bool_t hasShownCardErr = bool_false;
	bool_t hasShownUsb = bool_false;
	
	tosave_t toSave = 0;

    char fileName[8 + 1 + 3 + 1];

	while (1)
	{
        wdt_reset();
        
		USB_DoTask(); // this has to be taken care of in the main loop frequently
		// although it does nothing on the ATmega644 version
		
		if (alarmFlag == bool_true && snoozeFlag == bool_false)
		{            
            alarm();
            
            snoozeFlag = bool_true;

            mp3_powerDown(); // why here? because this will cause
            // playback to force resetting the volume back to what it should be

            // back to original settings
    
            hasShownCardErr = bool_false;
            hasShownUsb = bool_false;
            lcdContent = lcdshow_currentTime;
		}
		else if (USB_IsConnected() == 0)
		{
            hasShownUsb = bool_false;
            
            if (err != FR_OK || success == bool_false)
            {
                disk_initialize(0);
                err = f_mount(0, &fatfs);
                err = f_opendir(&dir, "/");
                                
                if (err == FR_OK)
                {
                    // card recently inserted or USB recently unplugged

                    lcd_setPos(0, 1);
                    //                           1234567890123456
                    fprintf_P(&lcd_stdout, PSTR("Loading Songs..."));
                    lcd_setPos(0, 2);
                    fprintf_P(&lcd_stdout, PSTR("                "));
                    wdt_reset();
                    songListInit(&dir);
                }

                if (err == FR_OK)
                {
                    if (ramSettings.playMode == playmode_shuffle)
                    {
                        lcd_setPos(0, 1);
                        //                           1234567890123456
                        fprintf_P(&lcd_stdout, PSTR("  Random Song:  "));
                        lcd_setPos(0, 2);
                        fprintf_P(&lcd_stdout, PSTR("                "));
                        lcd_setPos(0, 2);

                        success = openRandomFile(&dir, &fileInfo, &file, fileName);

                        //                           1234567890123456
                        fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                        lcdContent = lcdshow_other;
                        oneShotTimer_go(3);
                    }
                    else
                    {                        
                        lcd_setPos(0, 1);
                        //                           1234567890123456
                        fprintf_P(&lcd_stdout, PSTR(" Opening  Song: "));
                        lcd_setPos(0, 2);
                        fprintf_P(&lcd_stdout, PSTR("                "));
                        lcd_setPos(0, 2);

                        success = openNextFile(&dir, &fileInfo, &file, fileName);

                        //                           1234567890123456
                        fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                        lcdContent = lcdshow_other;
                        oneShotTimer_go(3);
                    }
                    
                    if (success != bool_false)
                    {
                        hasShownCardErr = bool_false;
                    }
                }
            }
            
            if (success == bool_false && hasShownCardErr == bool_false && isInMenu == bool_false)
            {
                mp3_powerDown();
                
                lcd_setPos(0, 1);
                //                           1234567890123456
                fprintf_P(&lcd_stdout, PSTR(" uSD Card Error "));
                lcd_setPos(0, 2);
                fprintf_P(&lcd_stdout, PSTR("                "));
                
                hasShownCardErr = bool_true;
                hasShownUsb = bool_false;
                isPlaying = bool_false;
                lcdContent = lcdshow_other;

                oneShotTimer_go(2);
            }
            else if (err == FR_OK)
            {
                hasShownCardErr = bool_false;
                
                if (isInMenu == bool_false)
                {
                    if (irctrl_getCmd() == ircmd_play && irctrl_isHeldDown() == 0)
                    {
                        // toggle play/pause
                        isPlaying = isPlaying == bool_false ? bool_true : bool_false;
                        irctrl_clearCmd();
                        
                        lcd_setPos(0, 1);
                        if (isPlaying == bool_false)
                        {
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR("  Paused Song:  "));

                            mp3_powerDown();
                        }
                        else
                        {
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR(" Playing  Song: "));
                        }
                        lcd_setPos(0, 2);
                        fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                        lcdContent = lcdshow_other;
                        oneShotTimer_go(3);
                    }
                    else if (irctrl_getCmd() == ircmd_stop && irctrl_isHeldDown() == 0)
                    {
                        // stop the song and rewind
                        isPlaying = bool_false;
                        mp3_powerDown();
                        f_lseek(&file, 0);
                        irctrl_clearCmd();
                        
                        lcd_setPos(0, 1);
                        //                           1234567890123456
                        fprintf_P(&lcd_stdout, PSTR(" Stopped  Song: "));
                        lcd_setPos(0, 2);
                        fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                        lcdContent = lcdshow_other;
                        oneShotTimer_go(3);
                    }
                    else if (irctrl_getCmd() == ircmd_left && irctrl_isHeldDown() == 0)
                    {
                        irctrl_clearCmd();
                        
                        // rewind, determine whether or not to rewind to start or go to previous song
                        if (mp3_getPlayTime() < 5)
                        {
                            // if only 5 seconds into the song, then previous song
                            
                            if (ramSettings.playMode == playmode_shuffle)
                            {
                                lcd_setPos(0, 1);
                                //                           1234567890123456
                                fprintf_P(&lcd_stdout, PSTR("  Random Song:  "));
                                lcd_setPos(0, 2);
                                fprintf_P(&lcd_stdout, PSTR("                "));
                                lcd_setPos(0, 2);

                                // respect the shuffle mode instead
                                success = openRandomFile(&dir, &fileInfo, &file, fileName);
                                mp3_resetPlayTime();

                                //                           1234567890123456
                                fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                                lcdContent = lcdshow_other;
                                oneShotTimer_go(3);
                            }
                            else
                            {
                                lcd_setPos(0, 1);
                                //                           1234567890123456
                                fprintf_P(&lcd_stdout, PSTR(" Previous Song: "));
                                lcd_setPos(0, 2);
                                fprintf_P(&lcd_stdout, PSTR("                "));
                                lcd_setPos(0, 2);
                                
                                success = openPrevFile(&dir, &fileInfo, &file, fileName);
                                mp3_resetPlayTime();

                                //                           1234567890123456
                                fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                                lcdContent = lcdshow_other;
                                oneShotTimer_go(3);
                            }
                        }
                        else
                        {
                            // past 5 seconds, rewind
                            
                            f_lseek(&file, 0);
                            mp3_resetPlayTime();
                            
                            lcd_setPos(0, 1);
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR("Rewinding  Song:"));
                            lcd_setPos(0, 2);
                            fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                            lcdContent = lcdshow_other;
                            oneShotTimer_go(3);
                        }
                        
                                                
                    }
                    else if (irctrl_getCmd() == ircmd_right && irctrl_isHeldDown() == 0)
                    {
                        irctrl_clearCmd();
                        
                        if (ramSettings.playMode == playmode_shuffle)
                        {
                            lcd_setPos(0, 1);
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR("  Random Song:  "));
                            lcd_setPos(0, 2);
                            fprintf_P(&lcd_stdout, PSTR("                "));
                            lcd_setPos(0, 2);

                            // respect the shuffle mode instead
                            success = openRandomFile(&dir, &fileInfo, &file, fileName);
                            mp3_resetPlayTime();
                        
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                            lcdContent = lcdshow_other;
                            oneShotTimer_go(3);
                        }
                        else
                        {
                            lcd_setPos(0, 1);
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR("   Next Song:   "));
                            lcd_setPos(0, 2);
                            fprintf_P(&lcd_stdout, PSTR("                "));
                            lcd_setPos(0, 2);

                            success = openNextFile(&dir, &fileInfo, &file, fileName);
                            mp3_resetPlayTime();
                        
                            //                           1234567890123456
                            fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                            lcdContent = lcdshow_other;
                            oneShotTimer_go(3);
                        }
                    }
                }
                
                if (isPlaying != bool_false)
                {
                    bool_t toMessWithVolume = (file.fptr == 0 || mp3_isPowered() == 0);

                    mp3_powerUp();

                    lastPlayStatus = mp3_playFile(&file);
                    if (lastPlayStatus == playstatus_endOfFile)
                    {
                        if (ramSettings.playMode == playmode_once)
                        {
                            isPlaying = bool_false;
                            f_lseek(&file, 0);
                            mp3_resetPlayTime();
                            mp3_powerDown();
                        }
                        else if (ramSettings.playMode == playmode_loop)
                        {
                            f_lseek(&file, 0);
                            mp3_resetPlayTime();
                        }
                        else if (ramSettings.playMode == playmode_shuffle)
                        {
                            if (lcdContent != lcdshow_menu)
                            {
                                lcd_setPos(0, 1);
                                //                           1234567890123456
                                fprintf_P(&lcd_stdout, PSTR("  Random Song:  "));
                                lcd_setPos(0, 2);
                                fprintf_P(&lcd_stdout, PSTR("                "));
                                lcd_setPos(0, 2);

                                lcdContent = lcdshow_other;
                            }
                            
                            success = openRandomFile(&dir, &fileInfo, &file, fileName);
                            mp3_resetPlayTime();
                            
                            if (lcdContent != lcdshow_menu)
                            {
                                fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);

                                oneShotTimer_go(3);
                            }
                        }
                        else
                        {
                            if (lcdContent != lcdshow_menu)
                            {
                                lcd_setPos(0, 1);
                                //                           1234567890123456
                                fprintf_P(&lcd_stdout, PSTR("   Next Song:   "));
                                lcd_setPos(0, 2);
                                fprintf_P(&lcd_stdout, PSTR("                "));
                                lcd_setPos(0, 2);

                                lcdContent = lcdshow_other;
                            }

                            success = openNextFile(&dir, &fileInfo, &file, fileName);
                            mp3_resetPlayTime();
                            
                            if (lcdContent != lcdshow_menu)
                            {
                                fprintf_P(&lcd_stdout, PSTR("  %s              "), fileName);
                                oneShotTimer_go(3);
                            }
                        }
                    }

                    if (success == bool_false)
                    {
                        mp3_powerDown();

                        lcd_setPos(0, 1);
                        //                           1234567890123456
                        fprintf_P(&lcd_stdout, PSTR(" uSD Card Error "));
                        lcd_setPos(0, 2);
                        fprintf_P(&lcd_stdout, PSTR("                "));

                        hasShownCardErr = bool_false;
                        isPlaying = bool_false;
                        lcdContent = lcdshow_other;
                        oneShotTimer_go(2);
                        err = FR_DISK_ERR;
                    }
                    else
                    {
                        if (toMessWithVolume != bool_false)
                        {
                            // putting this here fixes a bug
                            // the VS1033D does not change the volume unless sound
                            // has started playing and the previous volume is different
                            // from the one you are trying to set
                            if (ramSettings.mute == bool_false)
                            {
                                if (ramSettings.volume > 1)
                                {
                                    mp3_setVolume(ramSettings.volume - 1);
                                }
                                else
                                {
                                    mp3_setVolume(ramSettings.volume + 1);
                                }
                                mp3_setVolume(ramSettings.volume);
                            }
                            else
                            {
                                mp3_setVolume(2);
                                mp3_setVolume(1);
                            }
                        }
                    }
                }
            }
		}
        else if (USB_IsConnected() != 0 && hasShownUsb == bool_false && isUsbEnabled != bool_false)
        {
            // this should execute only once per USB plug-in event

            mp3_powerDown();

            if (lcdContent != lcdshow_menu)
            {
                hasShownUsb = bool_true;
                hasShownCardErr = bool_false;
                isPlaying = bool_false;

                lcd_setPos(0, 1);
                //                           1234567890123456
                fprintf_P(&lcd_stdout, PSTR("   USB   Mode   "));
                lcd_setPos(0, 2);
                fprintf_P(&lcd_stdout, PSTR("                "));

                lcdContent = lcdshow_other;
                oneShotTimer_go(7);
            }
        }

        // this should be processor free time, the decoder should be busy
        // and USB is already handled
		
		if (irctrl_getCmd() == ircmd_mute && irctrl_isHeldDown() == 0)
        {
            irctrl_clearCmd();
            
            if (ramSettings.mute == bool_false)
            {
                ramSettings.mute = bool_true;
                lcdContent = lcdContent != lcdshow_menu ? lcdshow_volume : lcdContent;
                mp3_setVolume(1);
            }
            else
            {
                ramSettings.mute = bool_false;
                lcdContent = lcdContent != lcdshow_menu ? lcdshow_volume : lcdContent;
                mp3_setVolume(ramSettings.volume);
            }

            oneShotTimer_go(1);
        }

        if (isInMenu == bool_false)
        {
            if (irctrl_getCmd() == ircmd_up)
            {
                irctrl_clearCmd();

                ramSettings.mute = bool_false;

                ramSettings.volume += VOLUME_STEP;
                if (ramSettings.volume > 255) ramSettings.volume = 255;

                lcdContent = lcdContent != lcdshow_menu ? lcdshow_volume : lcdContent;

                mp3_setVolume(ramSettings.volume);

                toSave |= tosave_ram;

                oneShotTimer_go(1);
            }
            else if (irctrl_getCmd() == ircmd_down)
            {
                irctrl_clearCmd();

                ramSettings.mute = bool_false;

                ramSettings.volume -= VOLUME_STEP;
                if (ramSettings.volume < 1) ramSettings.volume = 1; // do not set to 0

                lcdContent = lcdContent != lcdshow_menu ? lcdshow_volume : lcdContent;

                mp3_setVolume(ramSettings.volume);

                toSave |= tosave_ram;

                oneShotTimer_go(1);
            }
            else if (irctrl_getCmd() == ircmd_menu && irctrl_isHeldDown() == 0)
            {
                isInMenu = bool_true;
                irctrl_clearCmd();
                lcdContent = lcdshow_menu;
                oneShotTimer_go(5); // suppress time display
                hasShownCardErr = bool_false;
                toSave |= runMenu();
            }
        }
        else 
        {
            hasShownCardErr = bool_false;
            
            toSave |= runMenu();
        }

        bool_t hasTime = bool_true;

        if (toSave & tosave_eeprom && isInMenu == bool_false) // don't wear out eeprom too fast
        {
            hasTime = bool_false;
            saveToEeprom(&eepromSettings, sizeof(eepromSettings));
            toSave &= ~tosave_eeprom;
        }

        if (toSave & tosave_ram)
        {
            hasTime = bool_false;
            rtc_saveRam(&ramSettings, sizeof(ramSettings));
            toSave &= ~tosave_ram;
        }

        if (toSave & tosave_time)
        {
            hasTime = bool_false;
            uint8_t dayOfWeekConverted = currentTime.dayOfWeek + 1;
            rtc_setTimeBurst(dayOfWeekConverted, currentTime.hour, currentTime.minute, currentTime.second);
            toSave &= ~tosave_time;
        }

        if (hasTime != bool_false && updateRtc != bool_false)
        {
            // this is extra time consuming and we only need to do it at once per second at most
            updateRtc = bool_false;
            rtc_getTimeBurst(0, 0, 0, &(currentTime.dayOfWeek), &(currentTime.hour), &(currentTime.minute), &(currentTime.second));
            currentTime.dayOfWeek--; // rtc reports from 1-7, we want 0-6
        }
        else if (hasTime != bool_false)
        {            
            if (lcdContent == lcdshow_currentTime)
            {
                lcd_printTime(currentTime.hour, currentTime.minute, ramSettings.timeMode);
            }
            else if (lcdContent == lcdshow_volume)
            {
                lcd_setPos(0, 1);

                if (ramSettings.mute == bool_false)
                {
                    uint16_t vol = ramSettings.volume == 1 ? 0 : (ramSettings.volume == 255 ? 100 : ramSettings.volume * 100 / 255);

                    //                           1234567890123456
                    fprintf_P(&lcd_stdout, PSTR("  Volume:"));

                    // right align the number
                    if (vol < 10)
                    {
                        lcd_sendChar(' '); lcd_sendChar(' '); lcd_sendChar(' ');
                    }
                    else if (vol >= 100)
                    {
                        lcd_sendChar(' ');
                    }
                    else
                    {
                        lcd_sendChar(' '); lcd_sendChar(' ');
                    }

                    // print number with % symbol and blank rest-of-line
                    fprintf_P(&lcd_stdout, PSTR("%d%%  "), vol);
                    lcd_setPos(0, 2);

                    // print volume as a bar
                    int i;
                    for (i = 0; i < 16 && i < (vol * 16) / 100; i++)
                    {
                        lcd_sendChar(0xFF); // sends block characters
                    }
                    for (; i < 16; i++)
                    {
                        lcd_sendChar(' '); // blank rest of line
                    }
                }
                else
                {
                    //                           1234567890123456
                    fprintf_P(&lcd_stdout, PSTR("  Volume: Mute  "));
                    lcd_setPos(0, 2);
                    fprintf_P(&lcd_stdout, PSTR("                "));
                }
            }
            else if (lcdContent == lcdshow_menu)
            {
                showMenu();
            }
        }
	}
}