#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include "irctrl.h"
#include "ircodes.h"

/*
 * Do not ever enable IRCODE_USEFLASH
 * this IR command interpreter runs inside an ISR
 * through experimentation, I found that using codes stored in flash
 * memory caused the the function to run too slow for comfort
 *
 */

// global variables, volatile so they can be safely used inside interrupt routines
static volatile uint16_t pulses[IRCTRL_PULSE_NUMOFONPULSES * 2];
static volatile uint8_t currentOnPulseIdx;
static volatile uint16_t lastTimestamp;
static volatile uint8_t hasOverflowed;
static volatile ircmd_t lastCmd;
static volatile uint8_t heldFlag;

// reads the corresponding value stored in the code pattern memory
static inline uint16_t irctrl_getPattern(ircmd_t cmd, int idx)
{
    switch (cmd)
    {
        #ifdef IRCTRL_USEFLASH
        case ircmd_up:
            return pgm_read_word(&(ircmd_pattern_up[idx]));
        case ircmd_down:
            return pgm_read_word(&(ircmd_pattern_down[idx]));
        case ircmd_left:
            return pgm_read_word(&(ircmd_pattern_left[idx]));
        case ircmd_right:
            return pgm_read_word(&(ircmd_pattern_right[idx]));
        case ircmd_play:
            return pgm_read_word(&(ircmd_pattern_play[idx]));
        case ircmd_stop:
            return pgm_read_word(&(ircmd_pattern_stop[idx]));
        case ircmd_mute:
            return pgm_read_word(&(ircmd_pattern_mute[idx]));
        case ircmd_menu:
            return pgm_read_word(&(ircmd_pattern_menu[idx]));
        #else
        case ircmd_up:
            return ircmd_pattern_up[idx];
        case ircmd_down:
            return ircmd_pattern_down[idx];
        case ircmd_left:
            return ircmd_pattern_left[idx];
        case ircmd_right:
            return ircmd_pattern_right[idx];
        case ircmd_play:
            return ircmd_pattern_play[idx];
        case ircmd_stop:
            return ircmd_pattern_stop[idx];
        case ircmd_mute:
            return ircmd_pattern_mute[idx];
        case ircmd_menu:
            return ircmd_pattern_menu[idx];
        #endif
        default:
            return 0;
    }
}

// iterate through entire history backwards
// to find a matching command by process of elimination
// returns non-valid "ircmd_code_end" if no match found
// returns valid command code if match is found
ircmd_t ircmd_check()
{
    wdt_reset();
    uint8_t eliminated[ircmd_none];
    memset(eliminated, 0, ircmd_none); // clear this
    
    for (int i = 0, failedCnt = 0; i < currentOnPulseIdx && i < IRCTRL_PULSE_NUMOFONPULSES && failedCnt < ircmd_none; i++)
    { // for iterate through history of pulses
    
        for (ircmd_t k = 0; k < ircmd_none && failedCnt < ircmd_none; k++)
        { // for iterate through all commands
            
            if (eliminated[k] != 0)
            {
                // already eliminated, don't bother
                continue;
            }
            
            uint16_t shouldBeOnPulse = irctrl_getPattern(k, i * 2);
            uint16_t shouldBeOffPulse = irctrl_getPattern(k, (i * 2) + 1);
            
            if (pulses[i * 2] <= max((int16_t)(shouldBeOnPulse - IRCTRL_PULSE_ERROR), 0) ||
                pulses[i * 2] >= min((uint32_t)shouldBeOnPulse + IRCTRL_PULSE_ERROR, 0xFFFF) || 
                pulses[(i * 2) + 1] <= max((int16_t)(shouldBeOffPulse - IRCTRL_PULSE_ERROR), 0) ||
                pulses[(i * 2) + 1] >= min((uint32_t)shouldBeOffPulse + IRCTRL_PULSE_ERROR, 0xFFFF))
            {
                // out of range, eliminate
                eliminated[k] = 1;
                failedCnt++;
            }
        }
    }

    for (ircmd_t k = 0; k < ircmd_none; k++)
    { // for iterate through all commands, return the one not eliminated
        if (eliminated[k] == 0) return k;
    }

    return ircmd_none; // no match found, return non-valid result to indicate error
}

// output has changed, thus initiating a capture, causing this interrupt
ISR(IRCTRL_TIMER_CAPT_vect)
{
    // calculate using last timestamp
    uint32_t length = (((uint32_t)(IRCTRL_TIMER_ICR)) + 0xFFFF - lastTimestamp) & 0xFFFF;
    lastTimestamp = IRCTRL_TIMER_ICR;
    hasOverflowed = 0; // reset timer overflow flag
    
    if (bit_is_set(IRCTRL_PORTINPUT, IRCTRL_PIN))
    {
        // this is the end of an ON pulse
        
        if (length > IRCTRL_PULSE_STARTTHRESHOLD)
        {
            // command start
            currentOnPulseIdx = 0;
        }
        
        pulses[currentOnPulseIdx * 2] = length; // store, 0 index
        
        if (currentOnPulseIdx >= IRCTRL_PULSE_NUMOFONPULSES - 1)
        {
            ircmd_t cmd = ircmd_check();
            if (cmd != ircmd_none)
            {
                lastCmd = cmd;
                heldFlag = (heldFlag >= 1) ? 2 : 1;
            }
        }
        
        // toggle trigger edge
        IRCTRL_TIMER_TCCRxB &= ~_BV(IRCTRL_TIMER_ICIES);
    }
    else
    {
        pulses[(currentOnPulseIdx * 2) + 1] = length; // store
        
        // increment pointer with wrap-around
        currentOnPulseIdx = (currentOnPulseIdx + 1) % (IRCTRL_PULSE_NUMOFONPULSES + 1);
        
        // toggle trigger edge
        IRCTRL_TIMER_TCCRxB |= _BV(IRCTRL_TIMER_ICIES);
    }
}

// timer has overflown
ISR(IRCTRL_TIMER_OVF_vect)
{
    if (hasOverflowed != 0)
    {
        // this probably means buttons has been released
        
        // consecutive overflow event without output change, timeout
        pulses[(currentOnPulseIdx * 2) + 1] = 0xFFFF;
        
        heldFlag = heldFlag == 0 ? 0 : heldFlag - 1;
    }
    
    hasOverflowed = 1;
}

void irctrl_init()
{
    // initialize pin as input
    IRCTRL_PORTOUTPUT &= ~_BV(IRCTRL_PIN); // internal pullup disable (it's external)
    IRCTRL_PORTDIR &= ~_BV(IRCTRL_PIN);
    
    // completely reset
    hasOverflowed = 0;
    lastCmd = ircmd_none;
    currentOnPulseIdx = 0;
    for (int i = 0; i < IRCTRL_PULSE_NUMOFONPULSES * 2; i++)
    {
        pulses[i] = 0;
    }
    
    IRCTRL_TIMER_TCCRxB = 0b10000010; // start timer, enable input capture noise canceller, trigger on falling edge
    // prescaler is set to "div by 8"
    IRCTRL_TIMER_TIMSKx = _BV(IRCTRL_TIMER_ICIE) | _BV(IRCTRL_TIMER_TOIE); // enable timer capture interrupt and timer overflow interrupt
}

inline ircmd_t irctrl_getCmd()
{
    return lastCmd;
}

inline void irctrl_clearCmd()
{
    heldFlag = 0;
    lastCmd = ircmd_none;
}

inline uint8_t irctrl_isHeldDown()
{
    return heldFlag != 0;
}