#include "lcd.h"
#include <avr/io.h> // for AVR hardware
#include <avr/cpufunc.h> // for _NOP
#include <util/delay.h> // for delays
#include <avr/pgmspace.h> // for flash mem storage

// addresses for custom chars
#define upperlineaddr 0
#define ulc upperlineaddr
#define lowerlineaddr 1
#define llc lowerlineaddr
#define bothlineaddr 2
#define blc bothlineaddr
#define largecolonleftaddr 3
#define largecolonrightaddr 4
#define xxc 0xFF
#define ssc 0x20
#define upcharaddr 5
#define downcharaddr 6

// below are custom LCD characters, stored in flash

const uint8_t upperlinechar[8] PROGMEM = {
	0b00011111,
	0b00011111,
	0b00011111,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00000000
};

const uint8_t lowerlinechar[8] PROGMEM = {
	0b00000000,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00011111,
	0b00011111,
	0b00011111
};

const uint8_t bothlinechar[8] PROGMEM = {
	0b00011111,
	0b00011111,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00000000,
	0b00011111,
	0b00011111
};

const uint8_t largecolonleftchar[8] PROGMEM = {
	0b00000000,
	0b00000000,
	0b00000001,
	0b00000011,
	0b00000011,
	0b00000001,
	0b00000000,
	0b00000000
};

const uint8_t largecolonrightchar[8] PROGMEM = {
	0b00000000,
	0b00000000,
	0b00010000,
	0b00011000,
	0b00011000,
	0b00010000,
	0b00000000,
	0b00000000
};

const uint8_t upchar[8] PROGMEM = {
	0b00000000,
	0b00000100,
	0b00001110,
	0b00010101,
	0b00000100,
	0b00000100,
	0b00000100,
	0b00000000
};

const uint8_t downchar[8] PROGMEM = {
	0b00000000,
	0b00000100,
	0b00000100,
	0b00000100,
	0b00010101,
	0b00001110,
	0b00000100,
	0b00000000
};

const uint8_t numchar[10][6] PROGMEM = {
	{
		xxc, ulc, xxc,
		xxc, llc, xxc,
	},
	{
		ulc, xxc, ssc,
		llc, xxc, llc,
	},
	{
		ulc, blc, xxc,
		xxc, llc, llc,
	},
	{
		ulc, blc, xxc,
		llc, llc, xxc,
	},
	{
		xxc, llc, xxc,
		ssc, ssc, xxc,
	},
	{
		xxc, blc, ulc,
		llc, llc, xxc,
	},
	{
		xxc, blc, ulc,
		xxc, llc, xxc,
	},
	{
		ulc, ulc, xxc,
		ssc, ssc, xxc,
	},
	{
		xxc, blc, xxc,
		xxc, llc, xxc,
	},
	{
		xxc, blc, xxc,
		llc, llc, xxc,
	}
};

// initialize pin states
inline void lcd_initPins()
{
    LCD_DDR |= _BV(0) |
               _BV(1) |
               _BV(2) |
               _BV(3) |
               _BV(LCD_PIN_E) |
               _BV(LCD_PIN_RS) |
               _BV(LCD_PIN_LED);
    LCD_PORT = _BV(LCD_PIN_LED);
}

// send a byte to the LCD, RS pin must be setup before calling this function
static inline void lcd_sendByte(uint8_t data)
{
    // setup the data bits (first 4 bits)
    LCD_PORT = (LCD_PORT & 0xF0) | (0x0F & ((data & 0xF0) >> 4));
    _NOP(); _NOP(); _NOP(); // some delay
    
    // raise the E pin
    LCD_PORT |= _BV(LCD_PIN_E);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // lower the E pin, therefor inputting the data
    LCD_PORT &= ~_BV(LCD_PIN_E);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // setup the data bits (last 4 bits)
    LCD_PORT = (LCD_PORT & 0xF0) | (0x0F & data);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // raise the E pin
    LCD_PORT |= _BV(LCD_PIN_E);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // lower the E pin, therefor inputting the data
    LCD_PORT &= ~_BV(LCD_PIN_E);
    
    // minimum time for execution
    _delay_us(50);
}

// display a character on the LCD
void lcd_sendChar(char c)
{
    // displaying a character means we turn on RS
    LCD_PORT |= _BV(LCD_PIN_RS);
    
    lcd_sendByte(c);
}

// send a command to the LCD
void lcd_sendCmd(uint8_t data)
{
    // command means we turn off RS
    LCD_PORT &= ~_BV(LCD_PIN_RS);
    
    lcd_sendByte(data);
}

// sets the position of the cursor, x is 0 indexed, line is not
void lcd_setPos(uint8_t x, uint8_t line)
{
    if (line == 1)
    {
        lcd_sendCmd(0b10000000 | 0x00 | x);
    }
    else if (line == 2)
    {
        lcd_sendCmd(0b10000000 | 0x40 | x);
    }
    
    // TODO support for 4 line displays needs to be added
}

void lcd_setCustomChar(uint8_t address, uint8_t * dataArray)
{
	// send custom character definition
	uint8_t addr = address * 8;
	lcd_sendCmd(0b01000000 + addr);
	uint8_t i;
	for(i = 0; i < 8; i++)
	{
		lcd_sendChar(dataArray[i]);
	}
}

// initialize the LCD
void lcd_init()
{
    // setup the pins
    lcd_initPins();
    
    // clear the display
    lcd_sendCmd(0b00000001);
    _delay_us(1700); // requires a delay
    // return home
    lcd_sendCmd(0b00000010);
    _delay_us(1700); // requires a delay
    // these two effectively resets the display
    
    // function set, enable 4 bit mode first
    lcd_sendCmd(0b00101000);
    
    // entry mode set: left to right, no display shift
    //lcd_sendcmd(0b00000110); // commented out because it's already default
    
    // display control: on, no cursor, no cursor blink
    lcd_sendCmd(0b00001100);
    
    // set all custom fonts
    uint8_t temp[8];

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&upperlinechar[i]);
	}
	lcd_setCustomChar(upperlineaddr, temp);

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&lowerlinechar[i]);
	}
	lcd_setCustomChar(lowerlineaddr, temp);

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&bothlinechar[i]);
	}
	lcd_setCustomChar(bothlineaddr, temp);

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&largecolonleftchar[i]);
	}
	lcd_setCustomChar(largecolonleftaddr, temp);

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&largecolonrightchar[i]);
	}
	lcd_setCustomChar(largecolonrightaddr, temp);

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&upchar[i]);
	}
	lcd_setCustomChar(upcharaddr, temp);

	for(uint8_t i = 0; i < 8; i++)
	{
		temp[i] = pgm_read_byte(&downchar[i]);
	}
	lcd_setCustomChar(downcharaddr, temp);
    
    // start at line 1 position 0
    lcd_setPos(0, 1);
}

void lcd_printTime(uint8_t hour, uint8_t minute, uint8_t ampm)
{
    uint8_t a[4];
    uint8_t isPm = 0;

	// handle 24/12 hour modes
	if (hour == 0)
	{
		hour = 24;
	}

    if (hour >= 12)
    {
        isPm = 1;
    }

	if (ampm != 0)
	{
		hour %= 12;
		if(hour == 0)
		{
			hour = 12;
		}
	}

	// get individual digits
	a[0] = hour / 10;
	a[1] = hour % 10;
	a[2] = minute / 10;
	a[3] = minute % 10;

	// print characters
	for(uint8_t i = 0; i < 4; i++)
	{
		// handle spacing
		uint8_t j;
		if(i > 1)
		{
			j = 1;
		}
		else
		{
			j = 0;
		}

		uint8_t d = a[i];

		uint8_t e = (i * 4) + j;

		lcd_sendCmd(0b10000000 + e);
		lcd_sendChar(pgm_read_byte(&numchar[d][0]));
		lcd_sendChar(pgm_read_byte(&numchar[d][1]));
		lcd_sendChar(pgm_read_byte(&numchar[d][2]));
		lcd_sendCmd(0b11000000 + e);
		lcd_sendChar(pgm_read_byte(&numchar[d][3]));
		lcd_sendChar(pgm_read_byte(&numchar[d][4]));
		lcd_sendChar(pgm_read_byte(&numchar[d][5]));
	}
	
	// print colon and spacing
	if (ampm == 0)
	{
        lcd_sendCmd(0b10000000 + 4 + 3);
        lcd_sendChar(largecolonleftaddr);
        lcd_sendCmd(0b10000000 + 8);
        lcd_sendChar(largecolonrightaddr);
        
        lcd_sendCmd(0b11000000 + 4 + 3);
        lcd_sendChar(largecolonleftaddr);
        lcd_sendCmd(0b11000000 + 8);
        lcd_sendChar(largecolonrightaddr);
	}
	else if ((hour >= 0 && hour <= 11 || hour == 24) && isPm == 0)
	{
        lcd_sendCmd(0b10000000 + 4 + 3);
        lcd_sendChar('A');
        lcd_sendCmd(0b10000000 + 8);
        lcd_sendChar('M');
        
        lcd_sendCmd(0b11000000 + 4 + 3);
        lcd_sendChar(largecolonleftaddr);
        lcd_sendCmd(0b11000000 + 8);
        lcd_sendChar(largecolonrightaddr);
	}
	else if (isPm != 0)
	{
        lcd_sendCmd(0b10000000 + 4 + 3);
        lcd_sendChar(largecolonleftaddr);
        lcd_sendCmd(0b10000000 + 8);
        lcd_sendChar(largecolonrightaddr);
        
        lcd_sendCmd(0b11000000 + 4 + 3);
        lcd_sendChar('P');
        lcd_sendCmd(0b11000000 + 8);
        lcd_sendChar('M');
	}
	
	lcd_sendCmd(0b10000000 + 3);
	lcd_sendChar(' ');
	lcd_sendCmd(0b11000000 + 3);
	lcd_sendChar(' ');
	lcd_sendCmd(0b10000000 + 12);
	lcd_sendChar(' ');
	lcd_sendCmd(0b11000000 + 12);
	lcd_sendChar(' ');
}