#include "mp3.h"
#include <avr/io.h>
#include <avr/wdt.h>
#include <util/delay.h>

static uint8_t lastVolume;
static uint8_t volumeBeforeTest;
static uint8_t isPowered = 0;

inline void mp3_waitReady(uint8_t cnt)
{
    for(uint8_t i = 0; i < cnt; i++)
	{
		_delay_us(10);
		if (bit_is_clear(MP3_PORTINPUT, MP3_PIN_DREQ)) break;
	}
	loop_until_bit_is_set(MP3_PORTINPUT, MP3_PIN_DREQ);
}

inline uint8_t mp3_spi(uint8_t data)
{
    SPDR = data; // send
    loop_until_bit_is_set(SPSR, SPIF); // wait until sent
    return SPDR; // get reply
}

void mp3_writeReg(mp3_register_t addr, uint16_t data)
{
    spi_slow();
    
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_CS); // select
    
    mp3_spi(0b00000010); // write command
    mp3_spi(addr); // send address
    mp3_spi((data & 0xFF00) >> 8); // send data MSB first
    mp3_spi(data & 0x00FF);
    
    MP3_PORTOUTPUT |= _BV(MP3_PIN_CS); // deselect
    
    spi_fast();
    
    mp3_waitReady(10);
}

uint16_t mp3_readReg(mp3_register_t addr)
{
    spi_slow();
    
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_CS); // select
    
    mp3_spi(0b00000011); // write command
    mp3_spi(addr); // send address
    uint16_t read = mp3_spi(0) << 8; // read higher bits
    read |= mp3_spi(0); // read lower bits
    
    MP3_PORTOUTPUT |= _BV(MP3_PIN_CS); // deselect
    
    spi_fast();
    
    mp3_waitReady(10);
    
    return read;
}

// specify volume from 1-255 with 255 being loudest
// there's a math modification here because the volume adjustment is
// still non-linear
inline void mp3_setVolume(uint8_t volume)
{
    lastVolume = volume;
    int16_t deduct;
    if (volume == 255)
    {
        deduct = 0;
    }
    else if (volume == 254)
    {
        deduct = 1;
    }
    else if (volume <= 253 && volume >= 128)
    {
        deduct = (-volume / 2) + 128;
    }
    else if (volume <= 2)
    {
        deduct = 255 - volume;
    }
    else if (volume < 20)
    {
        deduct = ((-volume * 3) / 2) + 254;
    }
    else
    {
        deduct = ((-volume * 40) / 27) + 254;
    }

    if (deduct < 0) deduct = 0;
    if (deduct > 254) deduct = 254;

    mp3_writeReg(MP3_REG_VOL, ((deduct) << 8) | (deduct));
}

inline void mp3_resetPlayTime()
{
    mp3_writeReg(MP3_REG_DECODE_TIME, 0);
}

inline uint16_t mp3_getPlayTime()
{
    return mp3_readReg(MP3_REG_DECODE_TIME);
}

void mp3_startSineTest()
{
    volumeBeforeTest = lastVolume;
    
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_RST); // hardware reset
    _delay_us(10); // slight delay
    MP3_PORTOUTPUT |= _BV(MP3_PIN_RST); // bring out of reset, get ready

    mp3_waitReady(10);

    // enable test
    mp3_writeReg(MP3_REG_MODE, 0x0820); // see datasheet

    uint8_t testCmd[8] = { 0x53, 0xEF, 0x6E, 126, 0, 0, 0, 0, };
    mp3_playPacket(testCmd, 8); // send the test command
}

void mp3_stopSineTest()
{
    uint8_t testCmd[8] = { 0x45, 0x78, 0x69, 0x74, 0, 0, 0, 0, };
    mp3_playPacket(testCmd, 8); // send the test command

    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_RST); // hardware reset
    _delay_us(10); // slight delay
    MP3_PORTOUTPUT |= _BV(MP3_PIN_RST); // bring out of reset, get ready

    mp3_waitReady(10);

    // reset to defaults
    mp3_writeReg(MP3_REG_MODE, 0b0000100000000011); // see datasheet
    mp3_writeReg(MP3_REG_CLOCKF, 0x1800 | 0xE000); // see datasheet
    mp3_setVolume(volumeBeforeTest);
}

inline void mp3_playPacket(uint8_t* buffer, uint8_t length)
{
    spi_slow();
    loop_until_bit_is_set(MP3_PORTINPUT, MP3_PIN_DREQ); // wait for ready
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_DCS); // select
    for (uint8_t i = 0; i < length; i++)
    {
        mp3_spi(buffer[i]);
    }
    MP3_PORTOUTPUT |= _BV(MP3_PIN_DCS); // deselect
    spi_fast();
}

void mp3_init()
{
    // note that this must be called after the SD card has initialized to ensure SPI has initialized

    // setup all pin directions

    SPI_PORTDIR = (SPI_PORTDIR | (_BV(SPI_PIN_MOSI) | _BV(SPI_PIN_SCK) | _BV(SPI_PIN_SS))) & ~_BV(SPI_PIN_MISO);
    /* initialize SPI with lowest frequency; max. 400kHz during identification mode of card */
    SPCR = (0 << SPIE) | /* SPI Interrupt Enable Off */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (1 << SPR1) | /* Clock Frequency: f_OSC / 128 */
           (1 << SPR0);
    SPSR &= ~(1 << SPI2X); /* No doubled clock frequency */
    
    // reset pin, hold it low during setup
    MP3_PORTDIR |= _BV(MP3_PIN_RST);
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_RST);
    
    // DREQ is input, no pull-up
    MP3_PORTDIR &= ~_BV(MP3_PIN_DREQ);
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_DREQ);
    
    // both chip selects as output, starting high
    MP3_PORTDIR |= _BV(MP3_PIN_CS) | _BV(MP3_PIN_DCS);
    MP3_PORTOUTPUT |= _BV(MP3_PIN_CS) | _BV(MP3_PIN_DCS);
    
    _delay_us(10); // slight delay
    MP3_PORTOUTPUT |= _BV(MP3_PIN_RST); // bring out of reset, get ready
    
    mp3_waitReady(10);    
	
	mp3_writeReg(MP3_REG_MODE, 0b0000100000000011); // see datasheet
	mp3_writeReg(MP3_REG_CLOCKF, 0x1800 | 0xE000); // see datasheet

    isPowered = 1;
}

inline void mp3_powerUp()
{
    if (isPowered == 0)
    {
        isPowered = 1;
        mp3_init();
        mp3_setVolume(lastVolume);
    }
}

inline void mp3_powerDown()
{
    MP3_PORTOUTPUT &= ~_BV(MP3_PIN_RST); // power down
    //MP3_PORTOUTPUT &= ~_BV(MP3_PIN_CS); // no leakage current
    //MP3_PORTOUTPUT &= ~_BV(MP3_PIN_DCS); // no leakage current
    isPowered = 0;
}

inline uint8_t mp3_isPowered()
{
    return isPowered;
}